package ru.dbotthepony.mc.otm.capability.matter

import ru.dbotthepony.mc.otm.capability.FlowDirection
import ru.dbotthepony.mc.otm.capability.iterateProviders
import ru.dbotthepony.mc.otm.core.collect.map
import ru.dbotthepony.mc.otm.core.collect.reduce
import ru.dbotthepony.mc.otm.core.math.Decimal
import java.util.function.Supplier
import java.util.random.RandomGenerator

open class CombinedMatterStorage(
	final override val matterFlow: FlowDirection,
	val provider: Supplier<out Iterator<IMatterStorage>>,
	val random: Supplier<RandomGenerator?> = Supplier { null }
) : IMatterStorage {
	final override var storedMatter: Decimal
		get() = provider.get().map { it.storedMatter }.reduce(Decimal.ZERO, Decimal::plus)
		set(value) {
			throw UnsupportedOperationException()
		}
	final override val maxStoredMatter: Decimal
		get() = provider.get().map { it.maxStoredMatter }.reduce(Decimal.ZERO, Decimal::plus)

	final override fun receiveMatter(howMuch: Decimal, simulate: Boolean): Decimal {
		return provider.iterateProviders(random.get(), howMuch, simulate, IMatterStorage::receiveMatter)
	}

	final override fun extractMatter(howMuch: Decimal, simulate: Boolean): Decimal {
		return provider.iterateProviders(random.get(), howMuch, simulate, IMatterStorage::extractMatter)
	}

	final override val canSetMatterLevel: Boolean
		get() = false

	final override fun drainMatter(): Boolean {
		var any = false
		provider.get().forEach { any = it.drainMatter() || any }
		return any
	}

	final override fun fillMatter(): Boolean {
		var any = false
		provider.get().forEach { any = it.fillMatter() || any }
		return any
	}

	/*
	final override fun receiveMatterChecked(howMuch: Decimal, simulate: Boolean): Decimal {
		return provider.iterateProviders(random.get(), howMuch, simulate, IMatterStorage::receiveMatterChecked)
	}

	final override fun extractMatterChecked(howMuch: Decimal, simulate: Boolean): Decimal {
		return provider.iterateProviders(random.get(), howMuch, simulate, IMatterStorage::extractMatterChecked)
	}
	 */

	final override val missingMatter: Decimal
		get() = provider.get().iterator().map { it.missingMatter }.reduce(Decimal.ZERO, Decimal::plus)
}
